unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, ExtCtrls;

type
  TMainForm = class(TForm)
    TitleLabel: TLabel;
    MainStatus: TStatusBar;
    HealthScanTimer: TTimer;
    BDFPanel: TPanel;
    BusLabel: TLabel;
    DeviceLabel: TLabel;
    FunctionLabel: TLabel;
    BaselineGroup: TGroupBox;
    SaveButton: TButton;
    BaselineLabel: TLabel;
    BaselineScanTimer: TTimer;
    HealthGroup: TGroupBox;
    HealthLabel: TLabel;
    Button1: TButton;
    procedure FormCreate(Sender: TObject);
    procedure SaveButtonClick(Sender: TObject);
    procedure HealthScanTimerTimer(Sender: TObject);
    procedure BaselineScanTimerTimer(Sender: TObject);
  protected
    function TryLoadBaseline(Suffix: String): Boolean;
    procedure IdleScanForBus(Sender: TObject; var Done: Boolean);
    procedure IdleScanForBDF(Sender: TObject; var Done: Boolean);
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  Math,
  ACCES32;

const
  TarIDs = $0C57494F;
  SavedCaption = 'Saved.';

type
  TConfigSample = array[0..63] of LongWord;

  TRGBFColor = packed record
    case Boolean of
    True: (
      Color: TColor;
    );
    False: (
      Red: Byte;
      Green: Byte;
      Blue: Byte;
      Flags: Byte;
    );
  end;
  TPRGBFColor = ^TRGBFColor;

var
  ThisBDF: Word;
  iThisBus: Integer;
  Busses: array of Byte;
  BaselineData, BaselineMask: TConfigSample;
  ChangeAgo, ColorFresh: Integer;
  bBeenBad: Boolean;

function ConfigRead32(BDF: Word; Offset: LongWord): LongWord;
var
  BaseMessage: LongWord;

begin
  BaseMessage := $80000000 or (LongWord(BDF) shl 8) or (Offset and $FC);
  OutPortL($0CF8, BaseMessage);
  Result := InPortL($0CFC);
end;

procedure ReadConfigSample(BDF: Word; var IntoSample: TConfigSample);
var
  iReg: Integer;

begin
  for iReg := 0 to High(IntoSample) do
    IntoSample[iReg] := ConfigRead32(BDF, iReg shl 2)
  ;
end;

procedure DumpStringToFile(S, Path: String);
var
  Fil: TFileStream;

begin
  Fil := TFileStream.Create(Path, fmCreate or fmShareDenyWrite);
  try
    Fil.Write(S[1], Length(S));
  finally
    Fil.Free;
  end;
end;

function DumpFileToString(Path: String): String;
var
  Fil: TFileStream;
  L: Integer;

begin
  Fil := TFileStream.Create(Path, fmOpenRead or fmShareDenyNone);
  try
    L := Fil.Size;
    SetString(Result, nil, L);
    Fil.Read(Result[1], L);
  finally
    Fil.Free;
  end;
end;

procedure LoadDefaultFont(IntoFont: TFont);
var
  FontSize: Integer;
  NCM: TNonClientMetrics;

begin
  NCM.cbSize := SizeOf(NCM);
  if SystemParametersInfo(SPI_GETNONCLIENTMETRICS, SizeOf(NCM), @NCM, 0) then begin
    FontSize := Round(-3/4 * NCM.lfMessageFont.lfHeight);
    IntoFont.Name := NCM.lfMessageFont.lfFaceName;
  end
  else begin
    FontSize := Abs(IntoFont.Size);
    IntoFont.Name := 'Tahoma';
  end;
  IntoFont.PixelsPerInch := 72;
  IntoFont.Size := FontSize;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  LoadDefaultFont(Font);

  SaveButton.Align := alBottom;

  ThisBDF := 0;
  Application.OnIdle := IdleScanForBus;
  MainStatus.SimpleText := 'Scanning for busses';
end;

procedure TMainForm.IdleScanForBus(Sender: TObject; var Done: Boolean);
var
  iDevice, iBus: Integer;
  IDs: LongWord;

begin
  try
    Done := False;

    BusLabel.Caption := 'Bus: ' + IntToHex(ThisBDF shr 8, 2);
    DeviceLabel.Caption := 'Device: ' + '???';// IntToHex((ThisBDF shr 3) and $1F, 2);
    FunctionLabel.Caption := 'Function: ' + IntToHex(ThisBDF and $7, 1);
    for iDevice := 0 to $1F do begin
      IDs := ConfigRead32(ThisBDF or (iDevice shl 3), $00);
      if IDs <> $FFFFFFFF then begin
        iBus := Length(Busses);
        SetLength(Busses, iBus+1);
        Busses[iBus] := ThisBDF shr 8;
        Break;
      end;
    end;

    Inc(ThisBDF, 1 shl 8);
    if ThisBDF = 0 then begin
      if Length(Busses) <= 0 then begin
        Application.OnIdle := nil;
        MainStatus.SimpleText := 'Failed to find any PCI-type busses.';
        BusLabel.Caption := '';
        DeviceLabel.Caption := '';
        FunctionLabel.Caption := '';
      end
      else begin
        iThisBus := 0;
        ThisBDF := Busses[iThisBus] shl 8;
        Application.OnIdle := IdleScanForBDF;
        MainStatus.SimpleText := 'Found ' + IntToHex(Length(Busses), 1) + 'h busses, Scanning for devices';
      end;
    end;

  except
    on E: Exception do begin
      Application.OnIdle := nil;
      MainStatus.SimpleText := 'Error Halt: ' + E.ClassName + ': ' + E.Message;
    end;
  end;
end;

procedure TMainForm.IdleScanForBDF(Sender: TObject; var Done: Boolean);
var
  IDs: LongWord;
  iReg: Integer;

begin
  try
    Done := False;

    BusLabel.Caption := 'Bus: ' + IntToHex(ThisBDF shr 8, 2);
    DeviceLabel.Caption := 'Device: ' + IntToHex((ThisBDF shr 3) and $1F, 2);
    FunctionLabel.Caption := 'Function: ' + IntToHex(ThisBDF and $7, 1);
    IDs := ConfigRead32(ThisBDF, $00);
    if IDs = $FFFFFFFF then begin
    end
    else if IDs = TarIDs then begin
      ReadConfigSample(ThisBDF, BaselineData);
      for iReg := 0 to High(BaselineMask) do BaselineMask[iReg] := $FFFFFFFF;

      //if TryLoadBaseline(' Baseline.txt') then begin
        Application.OnIdle := nil;
        bBeenBad := False;
        ColorFresh := 0;

        {
        BaselineLabel.Caption := 'Loaded.';
        SaveButton.Visible := False;
        BaselineGroup.Visible := True;
        }

        HealthLabel.Caption := '';
        HealthGroup.Align := alClient;
        HealthGroup.Visible := True;
        HealthScanTimer.Enabled := True;
        MainStatus.SimpleText := 'Found device. Scanning health';
        HealthScanTimer.OnTimer(HealthScanTimer);
      {
      end
      else begin
        ReadConfigSample(ThisBDF, BaselineData);
        for iReg := 0 to High(BaselineMask) do BaselineMask[iReg] := $FFFFFFFF;
        TryLoadBaseline(' Seed.txt');
        Application.OnIdle := nil;
        ChangeAgo := 0;

        BaselineLabel.Caption := 'Scanning';
        SaveButton.Visible := True;
        BaselineGroup.Visible := True;
        BaselineScanTimer.Enabled := True;
        MainStatus.SimpleText := 'Found device, no baseline. Scanning for baseline';
      end;
      }
      Exit;
    end;
    Inc(ThisBDF, 1 shl 3);
    if (ThisBDF shr 8) <> Busses[iThisBus] then begin
      Inc(iThisBus);
      if iThisBus < Length(Busses) then
        ThisBDF := Busses[iThisBus] shl 8
      else begin
        Application.OnIdle := nil;
        MainStatus.SimpleText := 'Device not found.';
        BusLabel.Caption := '';
        DeviceLabel.Caption := '';
        FunctionLabel.Caption := '';
        Exit;
      end;
    end;

  except
    on E: Exception do begin
      Application.OnIdle := nil;
      MainStatus.SimpleText := 'Error Halt: ' + E.ClassName + ': ' + E.Message;
    end;
  end;
end;

procedure TMainForm.BaselineScanTimerTimer(Sender: TObject);
var
  NewSample: TConfigSample;
  iReg: Integer;
  Diff: LongWord;
  bChanged: Boolean;
  oColorFresh: Integer;
  AColor, BColor: TRGBFColor;
  Alpha, nAlpha: Double;

begin
  if not BaselineScanTimer.Enabled then Exit;

  oColorFresh := ColorFresh;

  ReadConfigSample(ThisBDF, NewSample);
  bChanged := False;
  for iReg := 0 to High(NewSample) do begin
    Diff := (NewSample[iReg] xor BaselineData[iReg]) and BaselineMask[iReg];
    if Diff = 0 then Continue;

    ChangeAgo := 0;
    ColorFresh := $100;
    bChanged := True;
    BaselineMask[iReg] := BaselineMask[iReg] and not Diff;
  end;

  if ColorFresh <> oColorFresh then begin
    //Set label color to strong contrast at full freshness.
    BColor.Color := GetSysColor(COLOR_BTNFACE);
    if (BColor.Red   and $80) <> 0 then AColor.Red   := $00 else AColor.Red   := $FF;
    if (BColor.Green and $80) <> 0 then AColor.Green := $00 else AColor.Green := $FF;
    if (BColor.Blue  and $80) <> 0 then AColor.Blue  := $00 else AColor.Blue  := $FF;
    AColor.Flags := 0;
    Alpha := ColorFresh/$100;
    nAlpha := 1 - Alpha;
    AColor.Red   := Round(AColor.Red   * Alpha + BColor.Red   * nAlpha);
    AColor.Green := Round(AColor.Green * Alpha + BColor.Green * nAlpha);
    AColor.Blue  := Round(AColor.Blue  * Alpha + BColor.Blue  * nAlpha);
    BaselineLabel.Color := AColor.Color;
    BaselineLabel.Font.Color := AColor.Color xor $FFFFFF;
  end;

  if bChanged then begin
    BaselineLabel.Caption := 'Changed!';
  end
  else begin
    if ColorFresh > 0 then ColorFresh := Max(0, ColorFresh-4);
    Inc(ChangeAgo);
    //Presumably a human will stop it before this happens, but just in case
    if ChangeAgo < 0 then begin
      BaselineScanTimer.Enabled := False;
      if BaselineLabel.Caption <> SavedCaption then
        BaselineLabel.Caption := 'Done.'
      ;
      Exit;
    end;
    if BaselineLabel.Caption <> SavedCaption then
      BaselineLabel.Caption := 'Steady ' + FloatToStrF(ChangeAgo/10, ffFixed, 10, 1) + 's'
    ;
  end;
end;

procedure TMainForm.SaveButtonClick(Sender: TObject);
var
  Baseline: TStringList;
  iReg, iBit: Integer;
  RegLine: String;

begin
  Baseline := TStringList.Create;

  RegLine := StringOfChar('?', 32);
  for iReg := 0 to High(BaselineData) do begin
    for iBit := 0 to 31 do
      if (BaselineMask[iReg] and (1 shl iBit)) = 0 then
        RegLine[1+iBit] := 'X'
      else if (BaselineData[iReg] and (1 shl iBit)) = 0 then
        RegLine[1+iBit] := '0'
      else
        RegLine[1+iBit] := '1'
    ;
    Baseline.Add(RegLine);
  end;

  Baseline.SaveToFile(IntToHex(TarIDs, 8) + ' Baseline.txt');
  BaselineLabel.Caption := SavedCaption;
  MainStatus.SimpleText := '';

  Baseline.Free;
end;

function TMainForm.TryLoadBaseline(Suffix: String): Boolean;
var
  FN: String;
  Baseline: TStringList;
  iReg, iBit, iCh: Integer;
  mBit: LongWord;
  RegLine: String;
  NewData, NewMask: TConfigSample;

begin
  try
    Result := False;
    FN := IntToHex(TarIDs, 8) + Suffix;
    if not FileExists(FN) then Exit;

    Baseline := TStringList.Create;
    try

      Baseline.LoadFromFile(FN);
      //Reserved for headers, including comments.
      while (Baseline.Count > 0) and (Baseline[0] <> '') and (Baseline[0][1] = '#') do
        Baseline.Delete(0)
      ;
      if Baseline.Count < Length(NewData) then Exit;

      for iReg := 0 to High(NewData) do begin
        RegLine := TrimLeft(Baseline[iReg]);
        if Length(RegLine) > 32 then begin
          iCh := Pos('#', RegLine);
          if iCh <> 0 then begin
            SetLength(RegLine, iCh-1);
            RegLine := TrimRight(RegLine);
          end;
          if Length(RegLine) > 32 then Exit;
        end;
        if Length(RegLine) < 32 then Exit;

        for iBit := 0 to 31 do begin
          mBit := 1 shl iBit;
          case RegLine[1+iBit] of
            '0': begin
              NewData[iReg] := NewData[iReg] and not mBit;
              NewMask[iReg] := NewMask[iReg] or mBit;
            end;
            '1': begin
              NewData[iReg] := NewData[iReg] or mBit;
              NewMask[iReg] := NewMask[iReg] or mBit;
            end;
            'X': begin
              NewMask[iReg] := NewMask[iReg] and not mBit;
            end;
            else Exit;
          end;
        end;
      end;

      for iReg := 0 to High(NewData) do BaselineData[iReg] := NewData[iReg];
      for iReg := 0 to High(NewMask) do BaselineMask[iReg] := NewMask[iReg];

    finally
      Baseline.Free;
    end;
  except
    Result := False;
  end;
end;

procedure TMainForm.HealthScanTimerTimer(Sender: TObject);
var
  NewSample: TConfigSample;
  iReg: Integer;
  bChanged: Boolean;

begin
  if not HealthScanTimer.Enabled then Exit;

  ReadConfigSample(ThisBDF, NewSample);
  bChanged := False;
  for iReg := 0 to High(NewSample) do
    if ((NewSample[iReg] xor BaselineData[iReg]) and BaselineMask[iReg]) <> 0 then begin
      bChanged := True;
      Break;
    end
  ;

  if bChanged then begin
    HealthLabel.Caption := 'Mismatch!';
    HealthLabel.Color := $0000FF;
    HealthLabel.Font.Color := $FFFFFF;

    bBeenBad := True;
  end
  else begin
    if bBeenBad then begin
      HealthLabel.Caption := 'Back To'#13#10'OK';
      HealthLabel.Color := $00FFFF;
      HealthLabel.Font.Color := $000000;
    end
    else begin
      HealthLabel.Caption := 'OK';
      HealthLabel.Color := $00FF00;
      HealthLabel.Font.Color := $000000;
    end;
  end;
end;

end.
